#!/usr/bin/env python
""" Plotting utilities for rocks."""
import numpy as np
import matplotlib.pyplot as plt


def scatter(catalogue, prop_name, nbins=10, show=False, savefig=""):
    """Create scatter/histogram figure for float parameters.

    Parameters
    ==========
    catalogue : rocks.core.propertyCollection
        A datacloud catalogue ingested in Rock instance.
    prop_name : str
        The property name, referring to a column in the datacloud table.
    nbins : int
        Number of bins in histogram. Default is 10
    show : bool
        Show plot. Default is False.
    save_to : str
        Save figure to path. Default is no saving.

    Returns
    =======
    matplotlib.figures.Figure instance, matplotib.axes.Axis instance
    """

    prop, errors = _property_errors(catalogue, prop_name)

    # ------
    # Build figure
    fig = plt.figure(figsize=(12, 8))
    gs = fig.add_gridspec(
        1,
        2,
        width_ratios=(7, 2),
        wspace=0.05,
        left=0.07,
        right=0.97,
        bottom=0.05,
        top=0.87,
    )

    ax = fig.add_subplot(gs[0])
    ax_histy = fig.add_subplot(gs[1], sharey=ax)

    avg, std = prop.weighted_average(errors, catalogue.preferred)
    ax.axhline(avg, label="Average", color=METHODS["avg"]["color"])
    ax.axhline(
        avg + std,
        label=fr"1$\sigma$ deviation",
        linestyle="dashed",
        color=METHODS["std"]["color"],
    )
    ax.axhline(avg - std, linestyle="dashed", color=METHODS["std"]["color"])

    x = np.linspace(1, len(prop), len(prop))
    for i, m in enumerate(np.unique(catalogue.method)):
        cur = np.where(np.asarray(catalogue.method) == m)
        fcol = "none"
        ax.scatter(
            x[cur],
            np.asarray(prop)[cur],
            label=m,
            marker=METHODS[m]["marker"],
            s=80,
            facecolors=fcol,
            edgecolors=METHODS[m]["color"],
        )
        ax.errorbar(
            x[cur],
            np.asarray(prop)[cur],
            yerr=np.asarray(errors)[cur],
            c=METHODS[m]["color"],
            linestyle="",
        )

    rejected = np.where(np.asarray(catalogue.preferred) == False)
    ax.scatter(
        x[rejected],
        np.asarray(prop)[rejected],
        label='Discarded', 
        marker='x',
        s=260,
        facecolors='gray',
    )

    ax.set_xticks(x)
    axtop = ax.twiny()
    axtop.set_xticks(x)
    axtop.set_xticklabels(catalogue.shortbib, rotation=25, ha="left")
    ax.set_ylabel(PLOTTING["LABELS"][prop_name])
    ax.legend(loc="best", ncol=2)

    range_ = ax.get_ylim()
    ax_histy.tick_params(axis="y", labelleft=False)
    ax_histy.hist(
        prop,
        bins=nbins,
        range=range_,
        orientation="horizontal",
        color="grey",
        label="All",
    )
    ax_histy.legend(loc="lower right")

    if savefig:
        fig.savefig(savefig)
        plt.close()

    if show:
        plt.show()
        plt.close()

    return fig, ax


def hist(catalogue, prop_name, nbins=10, show=False, save_to=""):
    """Create histogram figure for float parameters.

    Parameters
    ==========
    catalogue : rocks.core.propertyCollection
        A datacloud catalogue ingested in Rock instance.
    prop_name : str
        The property name, referring to a column in the datacloud table.
    nbins : int
        Number of bins in histogram. Default is 10
    show : bool
        Show plot. Default is False.
    save_to : str
        Save figure to path. Default is no saving.

    Returns
    =======
    matplotlib.figures.Figure instance, matplotib.axes.Axis instance
    """

    prop, errors = _property_errors(catalogue, prop_name)

    # ------
    # Build figure
    fig = plt.figure(figsize=(8, 6))

    plt.hist(prop, bins=nbins, label="Estimates")

    avg, std = prop.weighted_average(errors, catalogue.preferred)

    plt.errorbar(avg, 0.5, xerr=std, label="Average", marker="o")

    plt.legend(loc="upper right")

    plt.xlabel(PLOTTING["LABELS"][prop_name])
    plt.ylabel("Distribution")

    if save_to:
        fig.savefig(save_to)

    if show:
        plt.show()
        plt.close()

    return fig, plt.gca()


def _property_errors(catalogue, prop_name):
    """Retrieve main property and its errors from a datacloud catalogue.
    masses -> mass, diamalbedo -> either albedos or diameters

    Parameters
    ==========
    catalogue : rocks.core.propertyCollection
        Datacloud catalogue ingested into Rock instance.
    prop_name : str
        The property name, referring to a column in the datacloud table.

    Returns
    =======
    ndarray, ndarray
        The main property as defined in utils.DATACLOUD_META and its error.
    """
    prop = getattr(catalogue, prop_name)  # listSameTypeParameter

    if hasattr(catalogue, f"err_{prop_name}"):
        errors = getattr(catalogue, f"err_{prop_name}")
    else:
        errors = np.ones(np.array(prop).shape)

    return prop, errors


def show_scatter_hist(info, par, figname):

    # translate input
    n = len(info[1])
    x = np.linspace(1, n, n)

    val = info[1].get(par).values
    unc = info[1].get("err_" + par).values
    avg = info[0][0]
    std = info[0][1]

    # Define figure layout
    fig = plt.figure(figsize=(12, 8))
    gs = fig.add_gridspec(
        1,
        2,
        width_ratios=(7, 2),
        wspace=0.05,
        left=0.07,
        right=0.97,
        bottom=0.05,
        top=0.87,
    )
    ax = fig.add_subplot(gs[0])
    ax_histy = fig.add_subplot(gs[1], sharey=ax)

    # the scatter plot:
    # mean and std
    lavg = ax.axhline(avg, label="Average", color=s["avg"]["color"])
    lstd = ax.axhline(
        avg + std,
        label="1$\sigma$ deviation",
        color=s["std"]["color"],
        linestyle="dashed",
    )
    lstd = ax.axhline(avg - std, color=s["std"]["color"], linestyle="dashed")

    # all methods
    for i in np.unique(info[1].method):
        p = np.where(info[1].method == i)
        if True in info[1].selected.values[p]:
            fcol = s[i]["color"]
        else:
            fcol = "none"
        ax.scatter(
            x[p],
            val[p],
            marker=s[i]["marker"],
            s=80,
            label=i,
            facecolors=fcol,
            edgecolors=s[i]["color"],
        )
        ax.errorbar(x[p], val[p], yerr=unc[p], c=s[i]["color"], linestyle="")

    # axes
    ax.set_ylabel(PLOTTING["LABELS"][par])
    ax.set_xticks(x)

    # Legend
    ax.legend(loc="best", ncol=2)

    # place shortbib on top for quick identification
    axtop = ax.twiny()
    axtop.set_xticks(x)
    axtop.set_xticklabels(info[1].shortbib, rotation=25, ha="left")

    # Histogram
    range_ = ax.get_ylim()
    nbins = 10
    ax_histy.tick_params(axis="y", labelleft=False)
    sel = np.where(info[1].selected)

    na, ba, pa = ax_histy.hist(
        val,
        bins=nbins,
        range=range_,
        orientation="horizontal",
        color="grey",
        label="All",
    )
    ns, bs, ps = ax_histy.hist(
        val[sel],
        bins=nbins,
        range_=range_,
        orientation="horizontal",
        color="gold",
        label="Selected",
    )

    ax_histy.legend(loc="lower right")

    plt.savefig(figname)
    plt.close()


# Define colors/markers for all methods in ssodnet
METHODS = {
    "avg": {"color": "black"},
    "std": {"color": "darkgrey"},
    # -Space mission
    "SPACE": {"color": "gold", "marker": "X"},
    # -3d shape modeling
    "ADAM": {"color": "navy", "marker": "v"},
    "SAGE": {"color": "mediumblue", "marker": "^"},
    "KOALA": {"color": "slateblue", "marker": "<"},
    "Radar": {"color": "cornflowerblue", "marker": ">"},
    # -LC with scaling
    "LC+OCC": {"color": "lightgreen", "marker": "v"},
    "LC+AO": {"color": "forestgreen", "marker": "^"},  # deprecated =LC+IM
    "LC+IM": {"color": "forestgreen", "marker": "^"},
    "LC+TPM": {"color": "darkgreen", "marker": "<"},
    "LC-TPM": {"color": "green", "marker": ">"},
    # -Thermal models
    "STM": {"color": "grey", "marker": "D"},
    "NEATM": {"color": "grey", "marker": "o"},
    "TPM": {"color": "darkgrey", "marker": "s"},
    # -triaxial ellipsoid
    "TE-IM": {"color": "blue", "marker": "o"},
    "TE-Occ": {"color": "darkblue", "marker": "o"},
    # -2d on sky
    "OCC": {"color": "brown", "marker": "P"},
    "IM": {"color": "orange", "marker": "p"},
    "IM-PSF": {"color": "tomato", "marker": "H"},
    # -Mass from binary
    "Bin-IM": {"color": "navy", "marker": "v"},
    "Bin-Genoid": {"color": "mediumblue", "marker": "^"},
    "Bin-PheMu": {"color": "slateblue", "marker": "<"},
    "Bin-Radar": {"color": "cornflowerblue", "marker": ">"},
    # -Mass from deflection
    "DEFLECT": {"color": "brown", "marker": "D"},
    "EPHEM": {"color": "red", "marker": "o"},
    # -Taxonomy
    "Phot": {"color": "red", "marker": "s"},
    "Spec": {"color": "red", "marker": "s"},
}


PLOTTING = {
    "LABELS": {
        "diameter": "Diameter (km)",
        "mass": "Mass (kg)",
        "albedo": "Albedo",
    }
}



#!/usr/bin/env python
"""Utility functions for rocks."""

import collections
from functools import reduce
import json
import keyword
import os
import pickle
import sys
import time
import urllib
import warnings

import click
from iterfzf import iterfzf
import numpy as np
import pandas as pd
import requests
from tqdm import tqdm

import rocks


def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
    if "rocks/" in filename:
        return f"rocks: {message}\n"
    else:
        return "%s:%s: %s: %s\n" % (filename, lineno, category.__name__, message)


warnings.formatwarning = warning_on_one_line


def get_valid_ssocard_properties():
    if not os.path.isfile(rocks.PATH_TEMPLATE):
        rocks.utils.retrieve_ssoCard_template()

    with open(rocks.PATH_TEMPLATE, "r") as file_:
        TEMPLATE = json.load(file_)

    # ssoCard properties
    valid_props = [
        p for p in pd.json_normalize(TEMPLATE).columns if not p.startswith("datacloud")
    ]
    # missing intermediate levels
    valid_props = set(
        valid_props + [".".join(v.split(".")[:-1]) for v in valid_props if "." in v]
    )

    valid_props = sorted(list(valid_props), key=len)
    return valid_props


# ------
# Index functions
def read_index():
    """Read local index of asteroid numbers, names, SsODNet IDs.

    Returns
    =======
    pd.DataFrame
        Asteroid index with three columns.
    """
    with open(rocks.PATH_INDEX, "rb") as ind:
        return pickle.load(ind)


def create_index():
    """Update index of numbered SSOs."""

    # Get currently indexed objects
    index = read_index()

    # Get list of numbered asteroids from MPC
    url = "https://www.minorplanetcenter.net/iau/lists/NumberedMPs.txt"
    numbered = pd.read_fwf(url, colspecs=[(0, 7)], names=["number"])
    numbered = set(int(n.strip(" (")) for n in numbered.number)  # type: ignore

    # Compare list to index
    missing = set(index.number) ^ set(numbered)

    if not missing:
        return

    # Get ids of missing entries, append to index
    names, numbers, ids = zip(*rocks.identify(missing, return_id=True))

    index = index.append(
        pd.DataFrame(
            {
                "name": names,
                "number": numbers,
                "id_": ids,
            }
        )
    )

    # Save index to file
    index = (
        index.dropna(how="any")
        .drop_duplicates("number")
        .sort_values("number", ascending=True)
        .astype({"number": np.int64, "name": str, "id_": str})
    )

    index = index.reset_index()

    with open(rocks.PATH_INDEX, "wb") as ind:
        pickle.dump(index, ind, protocol=4)


def retrieve_index_from_repository():
    """Downloads the index of numbered asteroids from the rocks GitHub
    repository and saves it to the cache directory."""

    URL = "https://github.com/maxmahlke/rocks/blob/master/data/index.pkl?raw=True"

    index = pickle.load(urllib.request.urlopen(URL))
    index = index[["number", "name", "id_"]]

    with open(rocks.PATH_INDEX, "wb") as ind:
        pickle.dump(index, ind, protocol=4)


def select_sso_from_index():  # pragma: no cover
    """Select SSO numbers and designations from interactive fuzzy search.

    Returns
    =======
    str
        asteroid name or designation
    int
        asteroid number

    Notes
    -----
    If the selection is interrupted with ctrl-c, ``(None, None)`` is returned.

    Examples
    --------
    >>> import rocks
    >>> name, number, id_ = rocks.utils.select_sso_from_index()
    """
    INDEX = read_index()

    def _fuzzy_desig_selection(index):
        """Generator for fuzzy search of asteroid index file."""
        for number, name in zip(index.number, index["name"]):
            yield f"{number} {name}"

    try:
        nunaid = iterfzf(_fuzzy_desig_selection(INDEX), exact=True)
        number, *name = nunaid.split()

    except AttributeError:  # no SSO selected
        return None, np.nan, None

    id_ = INDEX.loc[INDEX.number == int(number), "id_"].values[0]

    return " ".join(name), int(number), id_


# ------
# ssoCard utility functions
# def rsetattr(obj, attr, val):  # pragma: no cover
#     """Deep version of setattr."""
#     pre, _, post = attr.rpartition(".")
#     return setattr(rgetattr(obj, pre) if pre else obj, post, val)
def rgetattr(obj, attr):
    """Deep version of getattr. Retrieve nested attributes.

    Parameters
    ==========
    obj
        Any python object with attributes.
    attr : str
        Attribut of obj. Nested attributes can be referrenced with '.'

    Returns
    =======
    _
        The requested attribute.
    """

    def _getattr(obj, attr):
        return getattr(obj, attr)

    return reduce(_getattr, [obj] + attr.split("."))


# def update_ssoCard(dict_, update):
#     """Recursively update ssoCard template with retrieved values.

#     Parameters
#     ==========
#     dict_ : dict, list
#         The values in the template or retrieved ssoCard.
#     update : dict, list
#         The values in the retrieved ssoCard.

#     Returns
#     =======
#     dict
#         The updated ssoCard.
#     """
#     for key, val in update.items():
#         if isinstance(val, collections.abc.Mapping):  # type: ignore
#             dict_[key] = update_ssoCard(dict_.get(key, {}), val)
#         else:
#             if isinstance(dict_, list):  # multiple pair members
#                 dict_ = dict_[-1]  # pragma: no cover
#             if isinstance(val, list):  # multiple taxonomies
#                 val = val[-1]
#             dict_[key] = val
#     return dict_


def retrieve_ssocard_template():
    """Retrieve current ssoCard template from SsODNet."""

    URL = (
        "https://ssp.imcce.fr/webservices/ssodnet/api/"
        "ssocard/templates/catalogue-template_aster-astorb.json"
    )

    response = requests.get(URL)

    if response.ok:
        ssoCard = response.json()

        with open(rocks.PATH_TEMPLATE, "w") as file_:
            json.dump(ssoCard, file_)

    else:
        warnings.warn(f"Retrieving the template failed with url:\n{URL}")


# def sanitize_keys(dict_):
#     """Recursively ensure that dict keys are not python keywords."""
#     for key, value in dict_.copy().items():
#         if isinstance(value, dict):
#             dict_[key] = sanitize_keys(dict_[key])
#         elif isinstance(value, list):
#             if all([isinstance(v, dict) for v in value]):
#                 dict_[key] = [sanitize_keys(v) for v in value]
#         if keyword.iskeyword(key):
#             dict_[f"{key}_"] = dict_.pop(key)
#     return dict_


# def set_endpoints_to_nan(dict_):
#     """Recursively sets dict values to NaN if they are not of type dict or unit."""
#     for key, value in dict_.copy().items():
#         if isinstance(value, dict):
#             dict_[key] = set_endpoints_to_nan(dict_[key])
#         else:
#             if key == "unit":
#                 continue
#             else:
#                 dict_[key] = np.nan
#     return dict_


# ------
# SsODNet functions
# def get_ssoCard(id_, only_cache=False):
#     """Retrieve single ssoCard either from cache or SsODNet.

#     Parameters
#     ==========
#     id_ : str
#         Single Minor body target id from SsODNet.
#     only_cache : bool
#         Do not query SsODNet, return None instead.

#     Returns
#     =======
#     dict, None
#         Single ssoCard in json format if successful.
#         None if query failed or not queried.
#     """
#     if not isinstance(id_, str):
#         warnings.warn(f"Expected string identifier, got {type(id_)}.")
#         return None

#     # Check presence in cache
#     PATH_CARD = os.path.join(rocks.PATH_CACHE, f"{id_}.json")

#     if not os.path.isfile(PATH_CARD):
#         if not only_cache:
#             return __query_ssoCards(id_)[0]
#         else:
#             return None
#     else:
#         with open(PATH_CARD, "r") as file_:
#             ssoCard = json.load(file_)
#         return ssoCard


# def get_ssoCards(ids, progress=False):
#     """Return target ssoCard. Use cache if existant, else, query SsODNet
#     and cache results. Accepts multiple ids.

#     Parameters
#     ==========
#     ids : str, list, np.array, pd.Series
#         Single Minor body target id from SsODNet or list of several.
#         Pass SsODNet ID for fast access.
#     progress : bool
#         Show progress of instantiation. Default is False.

#     Returns
#     =======
#     dict, list of dict, None
#         Single or list of ssoCards in json format if successful.
#         None if query failed.
#     """
#     if isinstance(ids, str):
#         ids = [ids]
#     elif isinstance(ids, pd.Series):
#         ids = ids.values

#     id_card = collections.OrderedDict(
#         (id_, get_ssoCard(id_, only_cache=True)) for id_ in ids
#     )
#     missing = [id_ for id_, card in id_card.items() if card is None]

#     if not missing:
#         return id_card.values()

#     print(f"Retrieving {len(missing)} ssoCards..")

#     # Perform POST query for missing cards in chunks
#     if progress:
#         progressbar = tqdm(desc="Retrieving ssoCards", total=len(missing))
#     else:
#         progressbar = False

#     for subset in [missing[i : i + 500] for i in range(0, len(missing), 500)]:

#         ssoCards = __query_ssoCards(subset, progressbar)

#         for id_, ssoCard in zip(subset, ssoCards):

#             if ssoCard is None:
#                 continue

#             # if ssoCard is None, this value is already None
#             id_card[id_] = ssoCard

#             with open(f"{rocks.PATH_CACHE}/{id_}.json", "w") as file_:
#                 json.dump(ssoCard, file_)

#     # Return requested ssoCards
#     return id_card.values()


# def __query_ssoCards(ids, progressbar=False):
#     """Retrieve ssoCards for targets from SsODNet via POST request.

#     Parameters
#     ==========
#     ids : str, list, np.array, pd.Series
#         Single Minor body target id from SsODNet or list of several.
#     progressbar : tqdm.tqdm
#         A progress bar instance. Default is False, no progressbar.

#     Returns
#     =======
#     list of dict, None
#         List of ssoCards in json format if successful.
#         None if query failed.

#     Notes
#     =====
#     Performs output check and tries to infer correct target string if query failed.
#     """
#     if isinstance(ids, str):
#         ids = [ids]
#     elif isinstance(ids, pd.Series):
#         ids = ids.values
#     if any(pd.isnull(ids)):

#         warnings.warn(
#             f"There {'are' if sum(pd.isnull(ids)) > 1 else 'is a'} NaN"
#             f" {'values' if sum(pd.isnull(ids)) > 1 else 'value'} among the IDs,"
#             f" returning None in {'their' if sum(pd.isnull(ids)) > 1 else 'its'} place."
#         )

#         # Get indices of NaN values
#         ind_nan = np.where(pd.isnull(ids))[0]

#         for ind in ind_nan:
#             ids.pop(ind)
#     else:
#         ind_nan = []

#     url = "https://asterws2.imcce.fr/webservices/ssodnet/api/ssocard/"

#     params = {"from": "rocks"}
#     file_ = {"targets": ("\n".join([id_ for id_ in ids]))}

#     try:
#         response = requests.post(url=f"{url}mytargets", params=params, files=file_)
#         response.encoding = "UTF-8"
#         response = response.json()
#     except requests.exceptions.RequestException as e:
#         warnings.warn(f"An error occurred when retrieving the ssoCards: {e}")
#         return None

#     ssoCards = [response[id_] for id_ in ids]

#     # Add False for the NaN indices removed above
#     for ind in ind_nan:
#         ssoCards.insert(ind, False)

#     if progressbar:
#         progressbar.update(n=len(ids))
#     return ssoCards


def get_unit(path_unit):
    """Get unit from units JSON file.

    Parameters
    ==========
    path_unit : str
        Path to the parameter in the JSON tree, starting at unit and
        separating the levels with periods.

    Returns
    =======
    str
        The unit of the requested parameter.
    """
    PATH_UNITS = os.path.join(
        os.path.expanduser("~"), "astro/rocks/data/unit-template_aster-astorb.json"
    )

    with open(PATH_UNITS, "r") as units:
        units = json.load(units)

    for key in path_unit.split("."):
        if key == "unit":
            unit = units[key]
        else:
            unit = unit[key]
    return unit


def get_description(name):
    PATH_UNITS = os.path.join(
        os.path.expanduser("~"), "astro/rocks/data/unit-template_aster-astorb.json"
    )

    breakpoint()

    print(pd.json_normalize(PATH_UNITS).columns)


# ------
# Pretty-print methods
def pretty_print_card(ssoCard, minimal):
    """Pretty-print ssoCard to terminal. Called by rocks info.

    Parameters
    ==========
    ssoCard : dict
        Minor body ssoCard.
    minimal : bool
        Reduce information to basics.
    """
    from rich.console import Console

    console = Console()

    if not minimal:
        # Print entire JSON
        console.print(ssoCard)
    else:

        from rich.panel import Panel
        from rocks.core import Rock

        rock = Rock(ssoCard["id"], ssoCard, skip_id_check=True)

        minimal_card = f"""

        type: {rock.type:<12} class: {rock.class_}
        taxonomy: {rock.taxonomy.class_:<12} ap: {rock.proper_elements.proper_semi_major_axis:.4}

        """  # {rock.proper_elements.semi_major_axis.unit}
        console.print(Panel(minimal_card, title=f"({rock.number}) {rock.name}"))


def pretty_print_property(rock, prop, plot=False):
    """Echo asteroid property for a single minor body.
    Print datacloud collections. Optionally open plots.

    Parameters
    ==========
    rock : rocks.Rock
        The asteroid Rock instance
    prop : str
        Asteroid property, attribute from JSON template
    plot : bool
        Plot propertyCollection. Default is False.
    """
    if isinstance(prop, rocks.core.stringParameter):
        _echo_stringParameter(prop)
    elif isinstance(prop, rocks.core.intParameter):
        _echo_intParameter(prop)
    elif isinstance(prop, rocks.core.floatParameter):
        _echo_floatParameter(prop)
    elif isinstance(prop, rocks.core.propertyCollection):
        _echo_propertyCollection(rock, prop, plot)
    elif isinstance(prop, rocks.core.listSameTypeParameter):
        _echo_listSameTyeParameter(prop, plot)
    else:
        print(prop)

    sys.exit()


def _echo_stringParameter(prop):
    """Echos stringParameter value."""
    print(prop)


def _echo_intParameter(prop):
    """Echos intParameter value."""
    if hasattr(prop, "unit"):
        unit = getattr(prop, "unit")
    else:
        unit = ""
    if hasattr(prop, "uncertainty"):
        uncert = getattr(prop, "uncertainty")
    else:
        uncert = ""

    if uncert:
        print(f"{prop:,} +- {uncert:,} {unit}")
    else:
        print(f"{prop}")


def _echo_floatParameter(prop):
    """Echos floatParameter value."""
    if hasattr(prop, "unit"):
        unit = getattr(prop, "unit")
    else:
        unit = ""
    if hasattr(prop, "uncertainty"):
        uncert = getattr(prop, "uncertainty")
    else:
        uncert = ""

    if uncert:
        print(f"{float(prop):.4} +- {float(uncert):.3} {unit}")
    else:
        print(f"{prop}")


def _echo_propertyCollection(rock, prop, plot):
    """Echos propertyCollection value.

    Parameters
    ==========
    rock : rocks.core.Rock
        The Rock instance.
    prop
        The attribute to print.
    """
    # TODO Add weighted average for floatParameters
    from rich import box
    from rich.table import Table
    from rich import print as rprint

    table = Table(
        header_style="bold blue",
        caption=f"({rock.number}) {rock.name}",
        box=box.SQUARE,
        footer_style="dim",
    )

    # Keys are table property names
    keys = [k for k in list(prop.__dict__.keys()) if k not in DONT_PRINT]

    for key in keys:
        table.add_column(key)

    if "diameters" in sys.argv:
        prop.preferred = prop.preferred_diameter
        prop_name = "diameter"
    elif "albedos" in sys.argv:
        prop.preferred = prop.preferred_albedo
        prop_name = "albedo"
    elif "masses" in sys.argv:
        prop_name = "mass"

    # Values are entries for each field
    for row in prop:
        table.add_row(
            *[str(getattr(row, key)) for key in keys],
            style="bold green"
            if hasattr(row, "preferred") and row.preferred
            else "white",
        )

    rprint(table)

    if plot:
        if any([p in sys.argv for p in ["-h", "--hist"]]):
            prop.hist(prop_name, show=True)
        if any([p in sys.argv for p in ["-s", "--scatter"]]):
            prop.scatter(prop_name, show=True)

    sys.exit()  # otherwise click prints Error


def _echo_listSameTyeParameter(prop):
    print(prop)


# ------
# Numerical methods
def weighted_average(observable, error):
    """Computes weighted average of observable.

    Parameters
    ----------
    observable : np.ndarray
        Float values of observable
    error : np.ndarray
        Corresponding errors of observable.

    Returns
    -------
    (float, float)
        Weighted average and its standard error.
    """
    if len(observable) == 1:
        return (observable[0], error[0])

    # Compute normalized weights
    weights = 1 / np.array(error) ** 2

    # Compute weighted average and uncertainty
    avg = np.average(observable, weights=weights)

    # Kirchner Case II
    # http://seismo.berkeley.edu/~kirchner/Toolkits/Toolkit_12.pdf
    var_avg = (
        len(observable)
        / (len(observable) - 1)
        * (
            sum(w * o ** 2 for w, o in zip(weights, observable)) / sum(weights)
            - avg ** 2
        )
    )
    std_avg = np.sqrt(var_avg / len(observable))
    return (avg, std_avg)


DONT_PRINT = [
    "num",
    "name",
    "id",
    "iddataset",
    "title",
    "url",
    "doi",
    "bibcode",
    "link",
    "datasetname",
    "idcollection",
    "selection",
    "source",
    "resourcename",
    "preferred",
    "preferred_albedo",
    "preferred_diameter",
    "_iter_index",
]



#!/usr/bin/env python
"""Implement the Datacloud catalogue pydantic models."""
from typing import List, Optional

import numpy as np
import pydantic
import rich

import rocks

# ------
# Validators
def ensure_list(value):
    """Ensure that parameters are always a list.

    Further replaces all None values by empty dictionaries.
    """
    if isinstance(value, (dict, int, float, str)):
        value = [value]

    return value


def ensure_int(value):
    return [int(v) for v in value]


# ------
# SsODNet catalogues as pydantic model
class Catalogue(pydantic.BaseModel):
    """The abstraction of a general datacloud catalogue on SsODNet."""

    id_: List[int] = pydantic.Field([np.nan], alias="id")
    number: List[int] = pydantic.Field([np.nan], alias="num")
    name: List[str] = [""]

    _ensure_int: classmethod = pydantic.validator(
        "id_", "number", allow_reuse=True, pre=True
    )(ensure_int)
    _ensure_list: classmethod = pydantic.validator("name", allow_reuse=True, pre=True)(
        ensure_list
    )


class Pairs(Catalogue):
    sibling_num: Optional[float] = np.nan
    sibling_name: Optional[str] = ""
    delta_v: Optional[float] = np.nan
    delta_a: Optional[float] = np.nan
    delta_e: Optional[float] = np.nan
    delta_sini: Optional[float] = np.nan
    delta_i: Optional[float] = np.nan
    membership: Optional[int] = None
    iddataset: Optional[str] = ""
    idcollection: Optional[int] = None
    resourcename: Optional[str] = ""
    datasetname: Optional[str] = ""
    bibcode: Optional[str] = ""
    title: Optional[str] = ""
    link: Optional[str] = ""


class AAMS(Catalogue):
    HG_H: Optional[float] = np.nan
    HG_G: Optional[float] = np.nan
    HG_r_err_H: Optional[float] = np.nan
    HG_l_err_H: Optional[float] = np.nan
    HG_r_err_G: Optional[float] = np.nan
    HG_l_err_G: Optional[float] = np.nan
    HG_rms: Optional[float] = np.nan
    HG_convergence: Optional[float] = np.nan
    HG1G2_H: Optional[float] = np.nan
    HG1G2_G1: Optional[float] = np.nan
    HG1G2_G2: Optional[float] = np.nan
    HG1G2_r_err_H: Optional[float] = np.nan
    HG1G2_l_err_H: Optional[float] = np.nan
    HG1G2_r_err_G1: Optional[float] = np.nan
    HG1G2_l_err_G1: Optional[float] = np.nan
    HG1G2_r_err_G2: Optional[float] = np.nan
    HG1G2_l_err_G2: Optional[float] = np.nan
    HG1G2_rms: Optional[float] = np.nan
    HG1G2_convergence: Optional[float] = np.nan
    HG12_H: Optional[float] = np.nan
    HG12_G12: Optional[float] = np.nan
    HG12_r_err_H: Optional[float] = np.nan
    HG12_l_err_H: Optional[float] = np.nan
    HG12_r_err_G12: Optional[float] = np.nan
    HG12_l_err_G12: Optional[float] = np.nan
    HG12_rms: Optional[float] = np.nan
    HG12_convergence: Optional[float] = np.nan
    HG_err_H: Optional[float] = np.nan
    HG_err_G: Optional[float] = np.nan
    HG1G2_err_H: Optional[float] = np.nan
    HG1G2_err_G1: Optional[float] = np.nan
    HG1G2_err_G2: Optional[float] = np.nan
    HG12_err_H: Optional[float] = np.nan
    HG12_err_G12: Optional[float] = np.nan
    iddataset: Optional[str] = ""
    idcollection: Optional[int] = None
    resourcename: Optional[str] = ""
    datasetname: Optional[str] = ""
    bibcode: Optional[str] = ""
    title: Optional[str] = ""
    link: Optional[str] = ""


class Astorb(Catalogue):
    OrbComputer: Optional[str] = ""
    H: Optional[float] = np.nan
    G: Optional[float] = np.nan
    B_V: Optional[float] = np.nan
    IRAS_Diameter: Optional[float] = np.nan
    IRAS_Class: Optional[str] = ""
    OrbArc: Optional[int] = None
    NumberObs: Optional[int] = None
    MeanAnomaly: Optional[float] = np.nan
    ArgPerihelion: Optional[float] = np.nan
    LongAscNode: Optional[float] = np.nan
    Inclination: Optional[float] = np.nan
    Eccentricity: Optional[float] = np.nan
    SemimajorAxis: Optional[float] = np.nan
    CEU_value: Optional[float] = np.nan
    CEU_rate: Optional[float] = np.nan
    PEU_value: Optional[float] = np.nan
    GPEU_fromCEU: Optional[float] = np.nan
    GPEU_fromPEU: Optional[float] = np.nan
    Note_1: Optional[int] = None
    Note_2: Optional[int] = None
    Note_3: Optional[int] = None
    Note_4: Optional[int] = None
    Note_5: Optional[int] = None
    Note_6: Optional[int] = None
    YY_calulation: Optional[int] = None
    MM_calulation: Optional[int] = None
    DD_calulation: Optional[int] = None
    YY_osc: Optional[int] = None
    MM_osc: Optional[int] = None
    DD_osc: Optional[int] = None
    CEU_yy: Optional[int] = None
    CEU_mm: Optional[int] = None
    CEU_dd: Optional[int] = None
    PEU_yy: Optional[int] = None
    PEU_mm: Optional[int] = None
    PEU_dd: Optional[int] = None
    GPEU_yy: Optional[int] = None
    GPEU_mm: Optional[int] = None
    GPEU_dd: Optional[int] = None
    GGPEU_yy: Optional[int] = None
    GGPEU_mm: Optional[int] = None
    GGPEU_dd: Optional[int] = None
    JD_osc: Optional[float] = np.nan
    px: Optional[float] = np.nan
    py: Optional[float] = np.nan
    pz: Optional[float] = np.nan
    vx: Optional[float] = np.nan
    vy: Optional[float] = np.nan
    vz: Optional[float] = np.nan
    MeanMotion: Optional[float] = np.nan
    OrbPeriod: Optional[float] = np.nan
    iddataset: Optional[str] = ""
    idcollection: Optional[int] = None
    resourcename: Optional[str] = ""
    datasetname: Optional[str] = ""
    bibcode: Optional[str] = ""
    title: Optional[str] = ""
    link: Optional[str] = ""


class AstDyS(Catalogue):
    H: Optional[float] = np.nan
    ProperSemimajorAxis: Optional[float] = np.nan
    err_ProperSemimajorAxis: Optional[float] = np.nan
    ProperEccentricity: Optional[float] = np.nan
    err_ProperEccentricity: Optional[float] = np.nan
    ProperSinI: Optional[float] = np.nan
    err_ProperSinI: Optional[float] = np.nan
    ProperInclination: Optional[float] = np.nan
    err_ProperInclination: Optional[float] = np.nan
    n: Optional[float] = np.nan
    err_n: Optional[float] = np.nan
    g: Optional[float] = np.nan
    err_g: Optional[float] = np.nan
    s: Optional[float] = np.nan
    err_s: Optional[float] = np.nan
    LCE: Optional[float] = np.nan
    My: Optional[float] = np.nan
    lam_fit: Optional[float] = pydantic.Field(np.nan, alias="lam-fit")
    iddataset: Optional[str] = ""
    idcollection: Optional[int] = None
    resourcename: Optional[str] = ""
    datasetname: Optional[str] = ""
    bibcode: Optional[str] = ""
    title: Optional[str] = ""
    link: Optional[str] = ""


class Taxonomies(Catalogue):
    link: List[str] = [""]
    datasetname: List[str] = [""]
    resourcename: List[str] = [""]
    doi: List[str] = [""]
    bibcode: List[str] = [""]
    url: List[str] = [""]
    title: List[str] = [""]
    source: List[str] = [""]
    shortbib: List[str] = [""]
    waverange: List[str] = [""]
    method: List[str] = [""]
    scheme: List[str] = [""]
    complex_: List[str] = pydantic.Field([""], alias="complex")
    class_: List[str] = pydantic.Field([""], alias="class")

    idcollection: List[Optional[int]] = [None]
    iddataset: List[Optional[int]] = [None]
    year: List[Optional[int]] = [None]

    preferred: List[bool] = [False]

    @pydantic.validator("preferred", pre=True)
    def select_preferred_taxonomy(cls, v, values):
        return rank_properties("taxonomy", values)

    def __len__(self):
        return len(self.class_)

    def __str__(self):

        if len(self) == 1 and not self.class_[0]:
            return "No taxonomies on record."

        table = rich.table.Table(
            header_style="bold blue",
            box=rich.box.SQUARE,
            footer_style="dim",
        )

        columns = ["class_", "complex_", "method", "waverange", "scheme", "shortbib"]

        for c in columns:
            table.add_column(c)

        # Values are entries for each field
        for i, preferred in enumerate(self.preferred):
            table.add_row(
                *[str(getattr(self, c)[i]) for c in columns],
                style="bold green" if preferred else "white",
            )
        rich.print(table)
        return ""


class Masses(Catalogue):
    link: List[str] = [""]
    datasetname: List[str] = [""]
    resourcename: List[str] = [""]
    doi: List[str] = [""]
    bibcode: List[str] = [""]
    url: List[str] = [""]
    title: List[str] = [""]
    source: List[str] = [""]
    shortbib: List[str] = [""]
    method: List[str] = [""]

    idcollection: List[Optional[int]] = [None]
    iddataset: List[Optional[int]] = [None]
    year: List[Optional[int]] = [None]
    mass: List[float] = [np.nan]
    err_mass: List[float] = [np.nan]

    preferred: List[bool] = [False]

    @pydantic.validator("preferred", pre=True)
    def select_preferred_mass(cls, v, values):
        return rank_properties("mass", values)

    def __len__(self):
        return len(self.mass)

    def __str__(self):

        if len(self) == 1 and not self.mass[0]:
            return "No masses on record."

        table = rich.table.Table(
            header_style="bold blue",
            box=rich.box.SQUARE,
            footer_style="dim",
        )

        columns = ["mass", "err_mass", "method", "shortbib"]

        for c in columns:
            table.add_column(c)

        # Values are entries for each field
        for i, preferred in enumerate(self.preferred):
            table.add_row(
                *[str(getattr(self, c)[i]) for c in columns],
                style="bold green" if preferred else "white",
            )
        rich.print(table)
        return ""


class Mpcatobs(Catalogue):
    link: List[str] = [""]
    datasetname: List[str] = [""]
    resourcename: List[str] = [""]
    doi: List[str] = [""]
    bibcode: List[str] = [""]
    title: List[str] = [""]
    iddataset: List[str] = [""]
    idcollection: List[Optional[int]] = [None]
    jd_obs: List[float] = [np.nan]
    ra_obs: List[float] = [np.nan]
    dec_obs: List[float] = [np.nan]
    mag: List[float] = [np.nan]
    vgs_x: List[float] = [np.nan]
    vgs_y: List[float] = [np.nan]
    vgs_z: List[float] = [np.nan]
    packed_name: List[str] = [""]
    orb_type: List[str] = [""]
    discovery: List[str] = [""]
    note1: List[str] = [""]
    note2: List[str] = [""]
    note3: List[str] = [""]
    note4: List[str] = [""]
    iau_code: List[str] = [""]
    date_obs: List[str] = [""]
    filter_: List[str] = pydantic.Field([""], alias="type")
    type_: List[str] = pydantic.Field([""], alias="type")


class Diamalbedo(Catalogue):

    link: List[str] = [""]
    datasetname: List[str] = [""]
    resourcename: List[str] = [""]
    doi: List[str] = [""]
    bibcode: List[str] = [""]
    url: List[str] = [""]
    title: List[str] = [""]
    iddataset: List[Optional[int]] = [None]
    method: List[str] = [""]
    shortbib: List[str] = [""]
    year: List[Optional[int]] = [None]
    source: List[str] = [""]

    idcollection: List[Optional[int]] = [None]
    albedo: List[float] = [np.nan]
    err_albedo: List[float] = [np.nan]
    diameter: List[float] = [np.nan]
    err_diameter: List[float] = [np.nan]
    beaming: List[float] = [np.nan]
    err_beaming: List[float] = [np.nan]
    emissivity: List[float] = [np.nan]
    err_emissivity: List[float] = [np.nan]

    preferred_albedo: List[bool] = [False]
    preferred_diameter: List[bool] = [False]

    @pydantic.validator("preferred_albedo", pre=True)
    def select_preferred_albedo(cls, v, values):
        return rank_properties("albedo", values)

    @pydantic.validator("preferred_diameter", pre=True)
    def select_preferred_diameter(cls, v, values):
        return rank_properties("diameter", values)

    def __len__(self):
        return len(self.albedo)

    def __str__(self):

        if len(self) == 1 and not self.albedo[0] and not self.diameter[0]:
            return "No diameters on record."

        table = rich.table.Table(
            header_style="bold blue",
            box=rich.box.SQUARE,
            footer_style="dim",
            caption="Blue: preferred diameters, yellow: preferred albedos, green: both preferred"
        )

        columns = [
            "albedo",
            "err_albedo",
            "diameter",
            "err_diameter",
            "method",
            "shortbib",
        ]

        for c in columns:
            table.add_column(c)

        # Values are entries for each field
        for i, preferred in enumerate(self.preferred_diameter):

            # blue for preferred diameter, yellow for preferred albedo, green for both,
            # else white
            if preferred:
                style = "bold green" if self.preferred_albedo[i] else "bold blue"
            elif self.preferred_albedo[i]:
                style = "bold yellow"
            else:
                style = "white"
            table.add_row(
                *[str(getattr(self, c)[i]) for c in columns],
                style=style,
            )
        rich.print(table)
        return ""


# ------
# Definitions
CATALOGUES = {
    # the catalogues as defined by rocks
    # rocks name :
    #     attr_name : Rock.xyz
    #     ssodnet_name : Name of catalogue in SsODNet
    "aams": {
        "attr_name": "aams",
        "ssodnet_name": "aams",
    },
    "albedos": {
        "attr_name": "diamalbedo",
        # "prop_name": {"diameters": "diameter", "albedos": "albedo"},
        "ssodnet_name": "diamalbedo",
    },
    "astdys": {
        "attr_name": "astdys",
        "ssodnet_name": "astdys",
    },
    "astorb": {
        "attr_name": "astorb",
        "ssodnet_name": "astorn",
    },
    "binarymp_tab": {
        "attr_name": "binaries",
        "ssodnet_name": "binarymp_tab",
    },
    "diamalbedo": {
        "attr_name": "diamalbedo",
        # "prop_name": {"diameters": "diameter", "albedos": "albedo"},
        "ssodnet_name": "diamalbedo",
    },
    "diameters": {
        "attr_name": "diamalbedo",
        # "prop_name": {"diameters": "diameter", "albedos": "albedo"},
        "ssodnet_name": "diamalbedo",
    },
    "families": {
        "attr_name": "families",
        "ssodnet_name": "families",
    },
    "masses": {
        "attr_name": "masses",
        # "prop_name": {"masses": "mass"},
        "ssodnet_name": "masses",
    },
    "mpcatobs": {
        "attr_name": "mpcatobs",
        "ssodnet_name": "mpcatons",
    },
    "pairs": {
        "attr_name": "pairs",
        "ssodnet_name": "pairs",
    },
    "taxonomies": {
        "attr_name": "taxonomies",
        "ssodnet_name": "taxonomy",
    },
}


def rank_properties(prop_name, obs):
    """Select ranking method based on property name.

    Parameters
    ==========
    prop_name : str
        The property to rank.
    obs : dict
        The observations including metadata


    Returns
    =======
    list of bool
        Entry "preferred" in propertyCollection, True if preferred, else False
    """
    if prop_name in ["taxonomy", "taxonomies"]:
        return select_taxonomy(obs)
    elif prop_name in ["mass", "masses"]:
        return select_numeric_property(obs, "mass")
    elif prop_name in ["albedo", "albedos"]:
        return select_numeric_property(obs, "albedo")
    elif prop_name in ["diameter", "diameters"]:
        return select_numeric_property(obs, "diameter")
    else:
        return [True for i in obs]


def select_taxonomy(taxonomies):
    """Select a single taxonomic classification from multiple choices.

    Evaluates the wavelength ranges, methods, schemes, and recency of
    classification.

    Parameters
    ----------
    taxonomies: dict
        Taxonomic classifications retrieved from SsODNet:datacloud.

    Returns
    -------
    list of bool
        True if preferred, else False
    """
    POINTS = {
        "scheme": {"bus-demeo": 3, "bus": 2, "smass": 2, "tholen": 1, "sdss": 1},
        "waverange": {"vis": 1, "nir": 3, "visnir": 6, "mix": 4},
        "method": {"spec": 7, "phot": 3, "mix": 4},
    }

    taxonomies = [dict(zip(taxonomies, t)) for t in zip(*taxonomies.values())]

    # Compute points of each classification
    points = []

    for row in taxonomies:
        points.append(
            sum(
                [
                    POINTS[crit][row[crit].lower()]
                    for crit in ["scheme", "waverange", "method"]
                ]
            )
        )

    # Convert points to boolean
    preferred = [True if p == max(points) else False for p in points]
    return preferred


def select_numeric_property(obs, prop_name):
    """Select preferred observations ranking methods.

    Parameters
    ==========
    obs : dict
        Property measurements and metadata retrieved from SsODNet:datacloud.
    prop_name : str
        Name of the asteroid property.

    Returns
    =======
    list of bool
        True if selected, else False.

    Notes
    =====
    The method ranking depends on the observable.
    """
    RANKING = PROPERTIES[prop_name]["ranking"]
    methods = set(obs["method"])

    for method in RANKING:

        if set(method) & methods:  # method used at least once

            # Ensure that rows do not contain all 0 values, as can be the case
            # for albedo in diamalbedo
            if all(
                [
                    obs[prop_name][i] == 0
                    for i, m in enumerate(obs["method"])
                    if m in method
                ]
            ):
                continue
            return [True if m in method else False for m in obs["method"]]

    return obs


# In alphabetic order
PROPERTIES = {
    # TEMPLATE
    # property_name: Rock instance attribute name
    #   attribute: attribute key in datacloud
    #   collection: Rock instance attribute name for collection of parameters
    #               Use plural form
    #   extra_columns: names of additional columns to output on CLI
    #   ssodnet_path: json path to asteroid property
    #                 (to be replaced by ssoCard
    #   type: asteroid property type, float or str
    "albedo": {
        "ranking": [
            ["SPACE"],
            ["ADAM", "KOALA", "SAGE", "Radar"],
            ["LC+TPM", "TPM", "LC+AO", "LC+Occ", "TE-IM", "TE-Occ"],
            ["AO", "Occ", "IM"],
            ["NEATM"],
            ["STM"],
        ],
    },
    "diameter": {
        "ranking": [
            ["SPACE"],
            ["ADAM", "KOALA", "SAGE", "Radar"],
            ["LC+TPM", "TPM", "LC+AO", "LC+Occ", "TE-IM", "TE-Occ"],
            ["AO", "Occ", "IM"],
            ["NEATM"],
            ["STM"],
        ],
    },
    "mass": {
        "ranking": [
            ["SPACE"],
            ["Bin-Genoid"],
            ["Bin-IM", "Bin-Radar", "Bin-PheMu"],
            ["EPHEM", "DEFLECT"],
        ],
    },
}


# Classes to complexes mapping
CLASS_TO_COMPLEX = {
    "A": "A",
    "Ad": "U",
    "AQ": "U",
    "AS": "U",
    "AU": "U",
    "AV": "U",
    "B": "B",
    "BC": "C",
    "BCF": "C",
    "BCU": "U",
    "BFC": "C",
    "BFU": "U",
    "BFX": "U",
    "Bk": "U",
    "BU": "B",
    "C": "C",
    "Caa": "C",
    "Cb": "C",
    "CB": "C",
    "CBU": "C",
    "CD": "U",
    "CDX": "U",
    "CF": "C",
    "CFB": "C",
    "CFU": "U",
    "CFXU": "U",
    "Cg": "C",
    "CG": "C",
    "CGSU": "U",
    "CGTP": "U",
    "CGU": "U",
    "Cgx": "C",
    "CL": "U",
    "CO": "U",
    "CP": "U",
    "CPF": "U",
    "CPU": "U",
    "CQ": "U",
    "CS": "U",
    "CSGU": "U",
    "CSU": "U",
    "CTGU": "U",
    "CU": "U",
    "CX": "U",
    "CXF": "U",
    "Cgh": "Ch",
    "Ch": "Ch",
    "D": "D",
    "DCX": "U",
    "DL": "D",
    "DP": "D",
    "DU": "D",
    "Ds": "D",
    "DS": "D",
    "DSU": "D",
    "DT": "D",
    "DTU": "D",
    "DU": "D",
    "DX": "D",
    "DXCU": "D",
    "E": "E",
    "EM": "X",
    "EU": "U",
    "F": "C",
    "FBCU": "U",
    "FC": "C",
    "FCB": "C",
    "FCU": "U",
    "FCX": "U",
    "FP": "U",
    "FU": "U",
    "FX": "U",
    "FXU": "U",
    "G": "C",
    "GC": "C",
    "GS": "U",
    "G": "U",
    "J": "V",
    "K": "K",
    "Kl": "U",
    "L": "L",
    "LA": "U",
    "Ld": "L",
    "LQ": "U",
    "LS": "U",
    "M": "M",
    "MU": "U",
    "O": "O",
    "OV": "U",
    "P": "P",
    "PC": "U",
    "PCD": "U",
    "PD": "U",
    "PDC": "U",
    "PF": "U",
    "PU": "U",
    "Q": "Q",
    "QO": "Q",
    "QRS": "U",
    "QSV": "U",
    "QV": "U",
    "Qw": "Q",
    "R": "R",
    "S": "S",
    "SA": "S",
    "Sa": "S",
    "SC": "U",
    "SCTU": "U",
    "SD": "U",
    "SDU": "U",
    "SG": "U",
    "Sk": "S",
    "Sl": "S",
    "SMU": "U",
    "SO": "S",
    "Sq": "S",
    "SQ": "S",
    "Sqw": "S",
    "Sr": "S",
    "SR": "S",
    "Srw": "S",
    "ST": "U",
    "STD": "U",
    "STGD": "U",
    "STU": "U",
    "SU": "U",
    "SV": "S",
    "Sv": "S",
    "Svw": "S",
    "Sw": "S",
    "SX": "U",
    "T": "T",
    "TCG": "U",
    "TD": "U",
    "TDG": "U",
    "TDS": "U",
    "TS": "U",
    "TSD": "U",
    "TX": "U",
    "V": "V",
    "Vw": "V",
    "X": "X",
    "XB": "U",
    "Xc": "X",
    "XC": "U",
    "XCU": "U",
    "XD": "U",
    "XDC": "U",
    "Xe": "X",
    "XF": "U",
    "XFC": "U",
    "XFCU": "U",
    "XFU": "U",
    "Xk": "X",
    "XL": "U",
    "Xn": "X",
    "XS": "U",
    "XSC": "U",
    "XSCU": "U",
    "XT": "U",
    "Xt": "X",
    "XU": "U",
    "Z": "U",
    np.nan: np.nan,
    None: np.nan,
    float("nan"): np.nan,
}



#!/usr/bin/env python
"""rocks command line suite."""
import json
import keyword
import os
import sys
import webbrowser

import click
import rich

import rocks


class AliasedGroup(click.Group):
    """Click group with custom default mode implementation."""

    def get_command(self, ctx, cmd_name):

        rv = click.Group.get_command(self, ctx, cmd_name)

        # ------
        # If it's a known subcommand, execute it
        if rv is not None:
            return rv

        # ------
        # Command aliases
        if cmd_name == "id":
            return identify

        # ------
        # Unknown subcommand -> echo asteroid property and optionally plot
        return echo()


@click.group(cls=AliasedGroup)
@click.version_option(version=rocks.__version__, message="%(version)s")
def cli_rocks():
    """CLI for minor body exploration."""
    pass


@cli_rocks.command()
def docs():
    """Open rocks documentation in browser."""
    webbrowser.open("https://rocks.readthedocs.io/en/latest/", new=2)


@cli_rocks.command()
def update():
    """Update index of numbered SSOs and SsODNet metadata.

    The list is retrieved from the Minor Planet Center, at
    https://www.minorplanetcenter.net/iau/lists/NumberedMPs.txt
    """
    rocks.utils.create_index()
    rocks.utils.retrieve_ssocard_template()


@cli_rocks.command()
def clear():
    """Clear the cached ssoCards."""
    for file_ in os.listdir(rocks.PATH_CACHE):
        if file_.endswith(".json") and file_ != "ssoCard_template.json":
            os.remove(os.path.join(rocks.PATH_CACHE, file_))


@cli_rocks.command()
@click.argument("this")
def identify(this):
    """Get asteroid name and number from string input.

    Parameters
    ==========
    this : str
        String to identify asteroid.
    """
    name, number = rocks.identify(this)  # type: ignore

    if isinstance(name, (str)):
        click.echo(f"({number}) {name}")


@cli_rocks.command()
@click.argument("this", default="")
@click.option(
    "-m", "--minimal", is_flag=True, help="Reduce output to basic information."
)
def info(this, minimal):
    """Print ssoCard of minor body.

    Parameters
    ==========
    this : str
        Minor body name, designation, or number.
        If empty, a selection is prompted.
    minimal : str
        Print a minimal overview of SSO properties. Default is False.
    """
    if not this:
        _, _, this = rocks.utils.select_sso_from_index()
    else:  # passed identified string, ensure that we know it
        _, _, this = rocks.identify(this, return_id=True)  # type: ignore

    if not isinstance(this, str):
        sys.exit()

    ssoCard = rocks.ssodnet.get_ssocard(this)

    if ssoCard is not None:
        rocks.utils.pretty_print_card(ssoCard, minimal)


@cli_rocks.command()
def properties():
    """Prints the ssoCard JSON keys and description using the template."""

    if not os.path.isfile(rocks.PATH_TEMPLATE):
        rocks.utils.retrieve_ssocard_template()

    with open(rocks.PATH_TEMPLATE, "r") as file_:
        TEMPLATE = json.load(file_)

    rich.print(TEMPLATE)


@cli_rocks.command()
def status():
    """Prints the availability of SsODNet:datacloud."""
    ceres = rocks.Rock(1)

    if hasattr(ceres, "taxonomy"):
        rich.print(r"[bold green]Datacloud is available.")
    else:
        print(r"[bold red]Datacloud is not available.")


def echo():
    """Echos asteroid property to command line. Optionally opens plot."""

    # Check if we're plotting the property
    plot = False

    for arg in ["-p", "--plot"]:

        if arg in sys.argv:
            sys.argv.remove(arg)

            plot = True

    # Get property and asteroid id
    _, prop, *id_ = sys.argv
    id_ = " ".join(id_)

    # Check if the property might be missing an underscore
    if keyword.iskeyword(prop):
        prop = f"{prop}_"

    if prop in rocks.datacloud.CATALOGUES.keys():
        datacloud = prop
    else:
        datacloud = []

    rock = rocks.Rock(id_, datacloud=datacloud)

    # Pretty-printing is implemented in the properties __str__
    rich.print(rocks.utils.rgetattr(rock, prop))
    sys.exit()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rocks']

package_data = \
{'': ['*']}

install_requires = \
['aiodns>=2.0.0,<3.0.0',
 'aiohttp>=3.7.3,<4.0.0',
 'cchardet>=2.1.7,<3.0.0',
 'chardet<4.0',
 'click>=7.1.2,<8.0.0',
 'iterfzf>=0.5.0,<0.6.0',
 'matplotlib>=3.3.4,<4.0.0',
 'numpy>=1.20.0,<2.0.0',
 'pandas>=1.2.1,<2.0.0',
 'requests>=2.25.1,<3.0.0',
 'rich>=9.10.0,<10.0.0',
 'tqdm>=4.56.0,<5.0.0']

extras_require = \
{'docs': ['sphinx>=3,<4',
          'sphinx-redactor-theme>=0.0.1,<0.0.2',
          'sphinx-click>=2.5.0,<3.0.0']}

entry_points = \
{'console_scripts': ['rocks = rocks.cli:cli_rocks']}

setup_kwargs = {
    'name': 'space-rocks',
    'version': '1.2.1',
    'description': 'Python client for SsODNet data access.',
    'long_description': '![PyPI](https://img.shields.io/pypi/v/space-rocks) [![Documentation Status](https://readthedocs.org/projects/rocks/badge/?version=latest)](https://rocks.readthedocs.io/en/latest/?badge=latest) [![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n# rocks\n\n*Disclaimer: The SsODNet service and its database are in an alpha version and under constant revision. The provided values and access methods may change without notice.*\n\n## Features\n\nCommand-line exploration and scripted retrieval of asteroid data.\n\n## Install\n\nInstall from PyPi using `pip`:\n\n     pip install space-rocks\n\n## Documentation\n\nCheck out the documentation at [read-the-docs.io](https://rocks.readthedocs.io/en/latest/).\n',
    'author': 'Max Mahlke',
    'author_email': 'max.mahlke@oca.eu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://rocks.readthedocs.io/en/latest/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0',
}


setup(**setup_kwargs)
